from collections.abc import Set
from datetime import datetime

import requests

from nicovideo_api_client.api.v2.field import SnapshotSearchAPIV2Fields
from nicovideo_api_client.api.v2.targets import SnapshotSearchAPIV2Targets
from nicovideo_api_client.constants import END_POINT_URL_V2_VERSION, FieldType, target_types


class SnapshotSearchAPIV2:
    """
    `ニコニコ動画 『スナップショット検索API v2』 <https://site.nicovideo.jp/search-api-docs/snapshot>`_
    のエンドポイント

    ここの関数から返るオブジェクトをメソッドチェーンをつないで呼び出すことで適切に設定された状態でAPIを叩くことができる。
    """

    @staticmethod
    def tags_exact() -> SnapshotSearchAPIV2Targets:
        """
        タグ検索を行う

        :return: クエリ(キーワード)入力オブジェクト
        """
        return SnapshotSearchAPIV2Targets(FieldType.TAGS_EXACT)

    @staticmethod
    def keywords() -> SnapshotSearchAPIV2Targets:
        """
        キーワード検索を行う

        :return: クエリ(キーワード)入力オブジェクト
        """
        return SnapshotSearchAPIV2Targets(FieldType.TITLE, FieldType.DESCRIPTION, FieldType.TAGS)

    @staticmethod
    def sing_voice_synthesis_tags() -> SnapshotSearchAPIV2Fields:
        """
        VOCALOID・UTAUなどの歌声合成ソフトを利用した動画を検索するタグを用いて検索する。
        タグの一覧は https://lit.link/avogado6 を参照。
        :return: レスポンスフィールドのタイプ指定オブジェクト
        """
        sing_voice_synthesis_tag = [
            "UTAU",
            "VOCALOID",
            "UTAUオリジナル曲",
            "VOCALOIDオリジナル曲",
            "VOICEROIDオリジナル曲",
            "NEUTRINOオリジナル曲",
            "CeVIOオリジナル曲",
            "vocaloid新曲リンク",
            "SynthesizerV",
            "歌うa.i.voice",
            "A.I.VOICEオリジナル曲",
            "歌うボイスロイド",
            "VOCALOIDインスト曲",
            "ボカロオリジナル曲",
            "ソフトウェアシンガー",
            "UTAU新曲リンク",
            "cevio新曲リンク",
            "VOCALOID処女作",
            "UTAU処女作",
            "CeVIO処女作",
            "neutrino(歌声合成エンジン)",
        ]
        exclude = ["ボカロオリジナルを歌ってみた"]
        return SnapshotSearchAPIV2Targets(FieldType.TAGS_EXACT).query(sing_voice_synthesis_tag, exclude)

    @staticmethod
    def targets(targets: Set[FieldType]) -> SnapshotSearchAPIV2Targets:
        """
        検索対象のフィールドタイプを指定して検索を行う。

        :return: クエリ(キーワード)入力オブジェクト
        """

        for t in targets:
            if t not in target_types:
                raise Exception("ターゲットに指定できないフィールドタイプです")

        if FieldType.TAGS_EXACT in targets and len(targets) > 1:
            raise Exception("tagsExact は他のフィールドタイプと併用して指定できません")
        return SnapshotSearchAPIV2Targets(*targets)

    @staticmethod
    def version(timeout: float = 400) -> datetime:
        """
        API の最終更新日時を取得する。

        `データの更新について <https://site.nicovideo.jp/search-api-docs/snapshot#
        %E3%83%87%E3%83%BC%E3%82%BF%E3%81%AE%E6%9B%B4%E6%96%B0%E3%81%AB%E3%81%A
        4%E3%81%84%E3%81%A6>`_ を参照。

        Note: 実際にリクエストが行われる。

        :param timeout: リクエストのタイムアウト時間。
        :return: 最終更新日時
        """
        json = requests.get(END_POINT_URL_V2_VERSION, timeout=timeout).json()
        return datetime.fromisoformat(json["last_modified"])



from typing import Dict, Optional, Set, Union

from nicovideo_api_client.api.v2.field import SnapshotSearchAPIV2Fields
from nicovideo_api_client.api.v2.sort import SnapshotSearchAPIV2Sort
from nicovideo_api_client.constants import FieldType


class SnapshotSearchAPIV2Targets:
    def __init__(self, *targets: FieldType):
        list_targets = list(targets)
        if len(list_targets) == 0:
            raise Exception("targets が設定されていません")
        self._query: Dict[str, str] = {"targets": ",".join(map(lambda x: x.value, list_targets))}

    def single_query(self, keyword: str) -> SnapshotSearchAPIV2Fields:
        """
        検索クエリ(キーワード)を指定する。

        `クエリ文字列仕様 <https://site.nicovideo.jp/search-api-docs/snapshot#%EF%BC%8A1-
        %E3%82%AF%E3%82%A8%E3%83%AA%E6%96%87%E5%AD%97%E5%88%97%E4%BB%95%E6%A7%98>`_
        に沿った値を入力することでキーワードを含めた検索を行うことができる。

        文字列を直接クエリに渡すため、キーワードに半角スペースやORを含めることでAND・OR検索を行うことができる。

        :param keyword: 検索するキーワード(文字列型)
        :return: レスポンスフィールドのタイプ指定オブジェクト
        """
        if type(keyword) is not str:
            raise TypeError("single_queryで指定するキーワードは str であるべきです")
        elif keyword == "":
            raise Exception("キーワードなし検索を行うにはno_keywordメソッドを指定する必要があります")
        self._query["q"] = keyword
        return SnapshotSearchAPIV2Fields(self._query)

    def query(self, keyword: list[str], exclude: Optional[list[str]] = None) -> "SnapshotSearchAPIV2And":
        """
        AND・OR検索を用いて検索クエリ(キーワード)を指定する。

        `クエリ文字列仕様 <https://site.nicovideo.jp/search-api-docs/snapshot#%EF%BC%8A1-
        %E3%82%AF%E3%82%A8%E3%83%AA%E6%96%87%E5%AD%97%E5%88%97%E4%BB%95%E6%A7%98>`_
        に沿った値を入力することで AND, OR 検索など複数のキーワードを含めた検索を行うことができる。

        :param exclude: NOT 検索で使用するキーワード
        :param keyword: 検索するキーワード。文字列を要素に持つリストの形式になっており、複数要素があれば OR で連結される。
        :return: レスポンスアンドのタイプ指定オブジェクト
        """

        if exclude is None:
            exclude = []
        if type(keyword) is not list or len(keyword) < 1:
            raise Exception("キーワードなし検索を行うにはno_keywordメソッドを指定する必要があります")
        return SnapshotSearchAPIV2And(self._query, exclude).and_(keyword)

    def no_keyword(self) -> SnapshotSearchAPIV2Fields:
        """
        キーワードなし検索を行う。

        `クエリ文字列仕様 <https://site.nicovideo.jp/search-api-docs/snapshot#%EF%BC%8A1-
        %E3%82%AF%E3%82%A8%E3%83%AA%E6%96%87%E5%AD%97%E5%88%97%E4%BB%95%E6%A7%98>`_
        q=自体の省略はできません。
        負荷の高い検索となりますので、filtersと併用しヒット件数を10万件以内に絞り込んだ上でご利用ください。

        :return: レスポンスフィールドのタイプ指定オブジェクト
        """
        self._query["q"] = ""
        return SnapshotSearchAPIV2Fields(self._query)


class SnapshotSearchAPIV2And(SnapshotSearchAPIV2Fields):
    def __init__(self, query: Dict[str, str], exclude: list[str]):
        super().__init__(query)
        self.exclude = exclude

    def field(self, fields: Set[FieldType]) -> SnapshotSearchAPIV2Sort:
        if self._query["q"] == "":
            raise Exception("NOT検索のみを実行することはできません")
        for e in self.exclude:
            self._arrange_keyword(e, exclude=True)
        return super().field(fields)

    def _arrange_keyword(self, keyword: str, exclude: bool = False):
        """
        キーワードを適切な形でクエリに指定する。
        :param keyword: クエリのqキーに指定するための文字列
        """
        # フレーズ検索かどうかの判定(" "は" OR "も含んでいる)
        if (
            ((keyword[0] != '"' and keyword[-1] != '"') and " " in keyword)
            or keyword[0] == "-"
            or '"' in keyword
            or "\\" in keyword
        ):
            keyword = '"' + keyword.replace("\\", "\\\\").replace('"', '\\"') + '"'
        elif keyword == "OR":
            keyword = '"OR"'
        if exclude:
            keyword = "-" + keyword
        if "q" not in self._query.keys():
            self._query["q"] = keyword
        else:
            self._query["q"] += " " + keyword

    def and_(self, keyword: Union[str, list[str]]) -> "SnapshotSearchAPIV2And":
        """
        AND 検索に利用する。
        NOT 検索に必要なパラメータは `query()` で指定する。
        :param keyword: 検索するキーワード。文字列または文字列を要素に持つリスト。リストにする場合は OR で連結される。
        """
        if type(keyword) is str:
            self._arrange_keyword(keyword)
        elif type(keyword) is list:
            for k in keyword:
                self._arrange_keyword(k)
                if k != keyword[-1]:
                    self._query["q"] += " OR"
        else:
            raise TypeError("検索キーワードには str または list が指定されるべきです")

        return self



import math
import time
from typing import Dict, List, Tuple
from urllib.parse import unquote_plus, urlencode

import requests

from nicovideo_api_client.api.v2.result import SnapshotSearchAPIV2Result
from nicovideo_api_client.constants import DEFAULT_RETRY, END_POINT_URL_V2


class SnapshotSearchAPIV2Request:
    def __init__(
        self,
        query: Dict[str, str],
        limit: int,
        user_agent: Tuple[str, str, str],
    ):
        self._query: Dict[str, str] = query
        self._limit = limit
        self._user_agent = user_agent

    def request(
        self,
        timeout: float = 400.0,
    ) -> SnapshotSearchAPIV2Result:
        """
        API にリクエストを送り、レスポンスを返す。

        `limit()` で 100件を超える値が設定されている場合、複数回のリクエストに分けて実行される。
        (そのため、 `timeout` の設定には注意が必要)

        :param timeout: リクエストを強制終了させるまでのタイムアウト時間。
        複数回に分割リクエストが行われる場合でも、リクエストの合計時間とこの値を比較する。
        また、`timeout` のうち 1/4 を接続時間に、 3/4 を読み込み時間に割り当てている。
        :return: レスポンスオブジェクト
        """

        if self._limit <= 100:
            self._query["_limit"] = str(self._limit)
        else:
            self._query["_limit"] = "100"

        total_time = 0.0

        (response, total_time) = self._request(timeout, total_time)

        results: List[SnapshotSearchAPIV2Result] = [response]

        total_count = int(results[0].total_count())

        if total_count <= self._limit:
            self._limit = total_count

        for pos in range(1, math.ceil(self._limit / 100)):
            self._query["_offset"] = str(pos * 100)
            if self._limit < (pos + 1) * 100:
                self._query["_limit"] = str(self._limit % 100)

            (response, total_time) = self._request(timeout, total_time)

            results.append(response)
        return SnapshotSearchAPIV2Result(self._query, results)

    def _request(
        self,
        timeout: float,
        total_time: float,
    ) -> Tuple[SnapshotSearchAPIV2Result, float]:
        # `'jsonFilter'` の値までまとめてエンコードされてしまっているので、それを避けるため True にしている
        # TODO: `'q'` がエンコードされずよくないので原因調査して修正する。
        url = self.build_url(True if "jsonFilter" in self._query else False)

        header = {"User-Agent": self._user_agent[0] + "/" + self._user_agent[1]}
        if self._user_agent[2] != "":
            header["User-Agent"] += " " + self._user_agent[2]

        for r in range(DEFAULT_RETRY):
            response = requests.get(url, timeout=(timeout / 4, timeout * 3 / 4), headers=header)
            response_time = response.elapsed.total_seconds()
            response_obj = SnapshotSearchAPIV2Result(self._query, response)
            total_time += response_time

            if total_time > timeout:
                raise TimeoutError("通信がタイムアウトしました")
            elif "meta" in response_obj.json() and response_obj.status()[0] == 200:
                break
            time.sleep(response_time)
        else:
            raise Exception("リトライ回数に達しました")
        time.sleep(response_time)
        tmp = SnapshotSearchAPIV2Result(self._query, response)
        return tmp, total_time

    def build_url(self, decode: bool = True) -> str:
        """
        リクエストオブジェクトに設定されたクエリパラメータから適切な
        URL 文字列を生成する。

        :param decode: True (デフォルト): urlencode を解除する, False: urlencode された文字列を返す
        :return: API リクエストのための URL 文字列
        """
        query = urlencode(self._query)
        if decode:
            query = unquote_plus(query)
        return END_POINT_URL_V2 + "?" + query



from typing import Dict, Union

from nicovideo_api_client.api.v2.request import SnapshotSearchAPIV2Request


class SnapshotSearchAPIV2UserAgent:
    def __init__(self, query: Dict[str, str], limit: int):
        self._query: Dict[str, str] = query
        self._limit: int = limit

    def user_agent(self, product: str, version: Union[int, str], comment: str = "") -> SnapshotSearchAPIV2Request:
        """
        レスポンスの要素数を指定する。

        :param product: HTTPリクエストヘッダのUser-Agentに指定するプロダクト名
        :param version: HTTPリクエストヘッダのUser-Agentに指定するプロダクトバージョン
        :param comment: HTTPリクエストヘッダのUser-Agentに指定するコメント
        :return: リクエストオブジェクト
        """

        if product.isspace():
            raise ValueError("User-Agentのプロダクト名に空白文字のみを指定することはできません")
        if version.isspace():
            raise ValueError("User-Agentのプロダクトバージョンに空白文字のみを指定することはできません")
        if version is int:
            version = str(version)
        user_agent = (product, version, comment)
        return SnapshotSearchAPIV2Request(self._query, self._limit, user_agent)
